#!/usr/bin/env python3

"""Copyright 2019 mnemonic AS <opensource@mnemonic.no>

Permission to use, copy, modify, and/or distribute this software for
any purpose with or without fee is hereby granted, provided that the
above copyright notice and this permission notice appear in all
copies.

THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL
WARRANTIES WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE
AUTHOR BE LIABLE FOR ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL
DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR
PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER
TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
PERFORMANCE OF THIS SOFTWARE.

---
Program to check the feed output directory of feed_downloader.py
Any downloaded content not found in the cache is handled.
The .meta files associated with the .html files from the feed download is
kept to preserve information about source and original titles as well as
feed time and date stamps.
This files builds a document understood by SCIOs 'doc' work queue.
This utility is not meant to handle the files downloaded from links _in_ the
feed as these document does not share any of the meta-data associated with
the feed entries. These files need to be handled separately (using SCIOs own
submit utility.
"""

import logging
import sys
from logging.handlers import RotatingFileHandler
from typing import Optional, Text


def setup_logging(
    loglevel: Text = "debug",
    logfile: Optional[Text] = None,
    prefix: Text = "scio",
    maxBytes: int = 100000000,  # 100 MB
    backupCount: int = 6,
) -> None:
    """Setup loglevel and optional log to file"""
    numeric_level = getattr(logging, loglevel.upper(), None)
    if not isinstance(numeric_level, int):
        raise ValueError("Invalid log level: %s" % loglevel)

    datefmt = "%Y-%m-%d %H:%M:%S"
    formatter = "[%(asctime)s] app=" + prefix + " level=%(levelname)s msg=%(message)s"

    if logfile:
        handlers = [
            RotatingFileHandler(logfile, maxBytes=maxBytes, backupCount=backupCount)
        ]

        logging.basicConfig(
            level=numeric_level, handlers=handlers, format=formatter, datefmt=datefmt
        )
    else:
        logging.basicConfig(
            level=numeric_level, stream=sys.stdout, format=formatter, datefmt=datefmt
        )



#!/usr/bin/env python3

""" SCIO Analyze module """

import argparse
import asyncio
import datetime
import gzip
import json
import logging
import os.path
import re
import sys
from typing import Dict, List, Optional, Text

import addict
import caep
import greenstalk
import pytz
import requests

import act.scio.config
import act.scio.logsetup
from act.scio import plugin

DEFAULT_METADATA_DATE_FIELDS = [
    "Creation-Date",
    "Last-Modified",
    "Last-Save-Date",
    "article:modified_time",
    "article:published_time",
    "citation_publication_date",
    "created",
    "date",
    "dcterms:created",
    "dcterms:modified",
    "meta:creation-date",
    "meta:save-date",
    "modified",
    "og:updated_time",
    "pdf:docinfo:created",
    "pdf:docinfo:custom:date",
    "pdf:docinfo:modified",
    "xmpMM:History:When",
]

ISO8601_DATE_RE = re.compile(r"^\d\d\d\d-\d\d-\d\dT\d\d:\d\d:\d\dZ$")


def parse_args() -> argparse.Namespace:
    """Helper setting up the argsparse configuration"""

    arg_parser = act.scio.config.parse_args("Scio 2 Analyzer")
    arg_parser.add_argument("--plugins", dest="plugins", type=str)
    arg_parser.add_argument(
        "--metadata-date-fields", default=",".join(DEFAULT_METADATA_DATE_FIELDS)
    )
    arg_parser.add_argument("--proxy-string", help="Proxy to use webdump upload")
    arg_parser.add_argument(
        "--webdump", dest="webdump", type=str, help="URI to post result data"
    )

    args = caep.config.handle_args(arg_parser, "scio/etc", "scio.ini", "analyze")

    args.metadata_date_fields = [
        field.strip() for field in args.metadata_date_fields.split(",")
    ]

    return args  # type: ignore


def remove_non_iso_dates(
    metadata: Dict[Text, Text], isodate_fields: List[Text]
) -> Dict[Text, Text]:
    filtered = {}

    for key, value in metadata.items():

        if key in isodate_fields:
            if not isinstance(value, str):
                logging.warning("date value is not string %s:%s", key, value)
                continue
            if not ISO8601_DATE_RE.search(value):
                logging.warning("Illegal date value in %s:%s", key, value)
                continue
        filtered[key] = value

    return filtered


def get_input(beanstalk_client: Optional[greenstalk.Client] = None) -> addict.Dict:
    """Helper function to abstract away how we get the text to work on"""

    nlpdata = addict.Dict()
    if not beanstalk_client:
        logging.info("Waiting for work on stdin")
        data = sys.stdin.read()
        nlpdata.content = data
    else:
        # ADD BEANSTALK JOB CONSUMPTION
        logging.info("Waiting for work from beanstalk")
        job = beanstalk_client.reserve()
        try:
            nlpdata = addict.Dict(json.loads(gzip.decompress(job.body)))
            logging.info("Started work on %s", nlpdata.get("hexdigest", "No Hexdigest"))
        except OSError as e:
            # File not found - log error
            logging.error(e)
        finally:
            # Remove job, either on success or file not found
            beanstalk_client.delete(job)

    return nlpdata


async def analyze(
    plugins: List[plugin.BasePlugin],
    beanstalk_client: Optional[greenstalk.Client] = None,
) -> addict.Dict:
    """Main analyze loop running all plugins on the text"""

    loop = asyncio.get_event_loop()

    nlpdata: addict.Dict = get_input(beanstalk_client)
    if not nlpdata.content:
        logging.error("Missing content")
        return addict.Dict({})

    nlpdata["Analyzed-Date"] = (
        datetime.datetime.utcnow().replace(tzinfo=pytz.UTC).isoformat()
    )

    # make sure we have a Creation-Date field even though the
    # document did not contain one. When missing, use current analyzed time.
    nlpdata["Creation-Date"] = nlpdata.get("metadata", {}).get(
        "Creation-Date", nlpdata["Analyzed-Date"]
    )

    staged = []  # for plugins with dependencies
    pipeline = []  # for plugins to be run now

    for p in plugins:
        if p.dependencies:
            staged.append(p)
        else:
            pipeline.append(p)

    while pipeline:
        tasks = []
        for p in pipeline:
            tasks.append(loop.create_task(p.analyze(nlpdata)))

        await asyncio.gather(*tasks, return_exceptions=True)

        for task in tasks:
            if task.exception():
                logging.error("%s returned an exception: %s", task, task.exception())
            else:
                res = task.result()
                nlpdata[res.name] = res.result

        pipeline = []
        for candidate in staged[:]:
            if all(dep in nlpdata for dep in candidate.dependencies):
                pipeline.append(candidate)
                staged.remove(candidate)

    for candidate in staged:
        logging.warning(
            "Candidate %s did not run due to unmet dependency %s",
            candidate.name,
            candidate.dependencies,
        )

    return nlpdata


async def async_main() -> None:
    """Async version of main"""

    args = parse_args()

    act.scio.logsetup.setup_logging(args.loglevel, args.logfile, "scio-analyze")

    plugins = plugin.load_default_plugins()

    if args.plugins:
        try:
            logging.info("loading plugins from %s", args.plugins)
            plugins += plugin.load_external_plugins(args.plugins)
        except FileNotFoundError:
            logging.warning("Unable to load plugins from %s", args.plugins)

    # Inject config directory into each plugin
    for p in plugins:
        p.configdir = os.path.join(args.config_dir, "etc/plugins")

    beanstalk_client = act.scio.config.beanstalk_client(args, watch="scio_analyze")
    elasticsearch_client = act.scio.config.elasticsearch_client(args)

    loop = asyncio.get_event_loop()

    while True:
        task = loop.create_task(analyze(plugins, beanstalk_client))
        try:
            await task
        except LookupError:
            logging.error(
                "Got LookupError. If nltk data is missing, "
                "run scio-nltk-download, which should download "
                "all nltk data to ~/nltk_data."
            )
            raise

        result = task.result()
        if not result:
            continue
        result_json = json.dumps(result, indent="  ")

        if args.webdump:
            proxies = (
                {"http": args.proxy_string, "https": args.proxy_string}
                if args.proxy_string
                else None
            )

            r = requests.post(args.webdump, data=result_json, proxies=proxies)
            if r.status_code != 200:
                logging.error("Unable to post result data to webdump: %s", r.text)

        if elasticsearch_client:
            hexdigest = result.get("hexdigest")

            if not hexdigest:
                logging.error("Missing hexdigest, skipping elasticsearch storage")
            else:
                result["metadata"] = remove_non_iso_dates(
                    result["metadata"], args.metadata_date_fields
                )

                try:
                    elasticsearch_client.index(index="scio2", id=hexdigest, body=result)
                except Exception as e:
                    logging.error("Error storing %s to elasticsearch: %s", hexdigest, e)
                    raise

                logging.info("Stored %s to elasticsearch", hexdigest)

        if not (args.webdump or elasticsearch_client):
            # Print to stdout if we do not send to webdump or elasticsearch
            print(result_json)

        # If we are not listening on a beanstalk work queue, behave like a command line
        # utility and exit after one document.
        if not beanstalk_client:
            break


def main() -> None:
    """Main entry point"""

    loop = asyncio.get_event_loop()
    task = loop.create_task(async_main())
    loop.run_until_complete(task)


if __name__ == "__main__":
    main()



"""All functions related to scio upload"""

import base64
import time
from logging import error, warning
from typing import IO, Dict, Text

import requests


def read_as_base64(obj: IO[bytes]) -> Text:
    """Create a base64 encoded string from a file like object"""

    encoded_bytes = base64.b64encode(obj.read())
    return encoded_bytes.decode("ascii")


def to_scio_submit_post_data(
    obj: IO[bytes], filemap: Dict[Text, Text], tlp: Text
) -> Dict[Text, Text]:
    """Take a file like object, and return a dictionary on the correct form for
    submitting to the SCIO API (https://github.com/mnemonic-no/act-scio-api)"""

    metadata = filemap.copy()
    metadata["content"] = read_as_base64(obj)
    metadata["tlp"] = tlp

    assert "content" in metadata
    assert "filename" in metadata
    assert "uri" in metadata
    assert "tlp" in metadata

    return metadata


def upload(url: Text, filemap: Dict[Text, Text], tlp: Text) -> None:
    """Upload a file to the Scio engine"""

    retry_codes = [429, None]

    with open(filemap["filename"], "rb") as file_h:
        post_data = to_scio_submit_post_data(file_h, filemap, tlp)

        # Disable all proxy settings from environment
        session = requests.Session()
        session.trust_env = False

        status_code = None
        while status_code in retry_codes:
            started = time.time()
            try:
                req = session.post(url, json=post_data)
            except requests.exceptions.ConnectionError as e:
                msg = (
                    f"Failed to upload to {url}. Time since upload started: "
                    + f"{time.time() - started} seconds. "
                    + f"Exception: {e}. post_data[filename]={post_data['filename']}, "
                    + f"post_data[uri]={post_data['uri']}"
                )
                error(msg)
                raise UploadError(msg)

            status_code = req.status_code

            if status_code not in retry_codes and status_code != 200:
                raise UploadError("Status {0}".format(req.status_code))

            if status_code in retry_codes:
                warning("%s: %s", status_code, req.text)
                time.sleep(1)


class UploadError(Exception):
    """Wrap all errors related to uploading data to scio"""



"""
Helper functions for retrieving vocabulary aliases
"""

import argparse
import re
from typing import Any, Dict, List, Optional, Text, Tuple

import requests
import urllib3

ESCAPE_CHARS = [",", ":", "#"]


def unescape(s: str) -> str:
    """
    Unescape
    """

    for char in ESCAPE_CHARS:
        s = s.replace("\\{}".format(char), char)

    return s


def escape(s: str) -> str:
    """
    Escape
    """
    for char in ESCAPE_CHARS:
        s = s.replace(char, "\\{}".format(char))

    return s


class UnknownResult(Exception):
    """UnknownResult is used in API request (not 200 result)"""

    def __init__(self, *args: Any, **kwargs: Dict[Any, Any]) -> None:
        Exception.__init__(self, *args, **kwargs)


def parse_aliases(line: str) -> Tuple[str, List[str]]:
    """
    Parse line on this form:

    name:alias1, alias2, alias3

    Supports escaping of ":", "," and "#"

    returns:
    name and list of aliases
    """

    # Split on ":" unless they are escaped
    vocab, aliases_str = re.split(r"(?<!\\):", line, maxsplit=1)

    # Split aliases on ",", unless they are escaped
    aliases = [
        unescape(a.strip()) for a in re.split(r"(?<!\\),", aliases_str) if a.strip()
    ]

    return unescape(vocab.strip()), aliases


def parseargs() -> argparse.ArgumentParser:
    """Parse arguments"""
    parser = argparse.ArgumentParser(description="Fetch ISO 3166 data")
    parser.add_argument(
        "--http-timeout", dest="timeout", type=int, default=120, help="Timeout"
    )
    parser.add_argument(
        "--proxy-string", dest="proxy_string", help="Proxy to use for external queries"
    )
    parser.add_argument(
        "--data-dir", dest="data_dir", required=True, help="Output path for data"
    )

    return parser


def output_alias(
    filename: str, alias_map: Dict[Text, Text], exclude: Optional[List[Text]] = None
) -> None:
    """
    Output alias to file
    """
    if not exclude:
        exclude = []

    with open(filename, "w") as f:
        for key in sorted(alias_map.keys()):
            aliases = alias_map[key]
            aliases_list = [
                escape(alias).strip()
                for alias in sorted(aliases)
                if alias not in exclude and alias != key
            ]
            f.write("{}:{}\n".format(escape(key).strip(), ",".join(aliases_list)))


def merge(
    *lists: Dict[Text, List[Text]], lower: bool = False
) -> Dict[Text, List[Text]]:
    """
    Combine multiple alias lists in order.
    If key exists as one of the previous alias lists, add aliases to the previous
    main key
    """

    combined: Dict[Text, List[Text]] = {}

    for alias_list in lists:
        for key, aliases in alias_list.items():
            if lower:
                key = key.lower()
                aliases = [alias.lower() for alias in aliases]

            # One of our aliases is already main group name in ATT&CK.
            # Merge aliases from MISP (including "main" group name with ATT&CK)
            alias_in_keys = [alias for alias in aliases if combined.get(alias)]

            if alias_in_keys:
                key = alias_in_keys[0]

            combined[key] = list(set(combined.get(key, []) + aliases))

    return combined


def fetch_json(
    url: str, proxy_string: str = "", timeout: int = 60, verify_https: bool = True
) -> Dict[Text, Any]:
    """Fetch remote URL as JSON
    url (string):                    URL to fetch
    proxy_string (string, optional): Optional proxy string on format host:port
    timeout (int, optional):         Timeout value for query (default=60 seconds)
    """

    proxies = {"http": proxy_string, "https": proxy_string}

    options = {
        "verify": verify_https,
        "timeout": timeout,
        "proxies": proxies,
        "params": {},
    }

    if not verify_https:
        urllib3.disable_warnings(urllib3.exceptions.InsecureRequestWarning)  # type: ignore

    # Unable to infer type
    req = requests.get(url, **options)  # type: ignore

    if not req.status_code == 200:
        errmsg = "status_code: {0.status_code}: {0.content}"
        raise UnknownResult(errmsg.format(req))

    # Unable to infer type of json()
    return req.json()  # type: ignore
